import random as rnd

EMPTY = ' '
DEAD = 'X'
HIT = '+'
MISSED = '-'
SHIP = 'O'

LETTERKEYS = [
    'A',
    'B',
    'C',
    'D',
    'E',
    'F',
    'G',
    'H',
    'I',
    'J'
]


def digit(key):
    if key in LETTERKEYS:
        return LETTERKEYS.index(key) + 1
    elif 1 <= key <= 10:
        return key
    else:
        raise ValueError


def letter(key):
    if key in LETTERKEYS:
        return key
    elif 1 <= key <= 10:
        return LETTERKEYS[key-1]
    else:
        raise ValueError


def area(posx, posy):
    res = []
    for i in range(3):
        for j in range(3):
            res.append((posx - 1 + i, posy - 1 + j))
    return res


def around(posx, posy):
    return [a for a in area(posx, posy) if
            not (a[0] == posx and a[1] == posy) and (1 <= a[0] <= 10 and 1 <= a[1] <= 10)]


def cross_around(posx, posy):
    res = []
    for i in range(2):
        res.append((posx, posy - 1 + i * 2))
        res.append((posx - 1 + i * 2, posy))
    return [a for a in res if (1 <= a[0] <= 10 and 1 <= a[1] <= 10)]


# Класс-родитель корабля
class SBGameShip:
    def __init__(self, field, angle = None, length = None, posx = None, posy = None):
        self.angle = 0 if angle is None else angle
        self.length = 0 if length is None else length
        self.posx = 0 if not posx else posx
        self.posy = 0 if not posy else posy
        self.field = field
        self.cells = list()

    def __randomcoords(self):
        posx = 0
        posy = 0
        self.angle = rnd.randint(0, 3)
        if self.angle == 0:
            posx = rnd.randint(self.length, self.field.xlength)
            posy = rnd.randint(1, self.field.ylength)
        elif self.angle == 1:
            posx = rnd.randint(1, self.field.xlength)
            posy = rnd.randint(1, self.field.ylength - self.length)
        elif self.angle == 2:
            posx = rnd.randint(1, self.field.xlength - self.length)
            posy = rnd.randint(1, self.field.ylength)
        elif self.angle == 3:
            posx = rnd.randint(1, self.field.xlength)
            posy = rnd.randint(self.length, self.field.ylength)

        self.posx = posx
        self.posy = posy

        self.gen_cells()

    def gen_cells(self):
        if self.angle == 0:
            self.cells = [[self.posx - cell, self.posy, SHIP] for cell in range(self.length)]
        elif self.angle == 1:
            self.cells = [[self.posx, self.posy + cell, SHIP] for cell in range(self.length)]
        elif self.angle == 2:
            self.cells = [[self.posx + cell, self.posy, SHIP] for cell in range(self.length)]
        elif self.angle == 3:
            self.cells = [[self.posx, self.posy - cell, SHIP] for cell in range(self.length)]

    def randompos(self):
        ship_not_put = True
        while (ship_not_put):
            try:
                self.__randomcoords()
                self.field.put_ship(self)
                ship_not_put = False
            except ValueError:
                pass

    def isdead(self):
        is_d = False not in [a[2] == HIT for a in self.cells]
        if is_d:
            for i, cell in enumerate(self.cells):
                self.cells[i][2] = DEAD
        return is_d

    def around(self):
        res = set()
        own_cells = [(cell[0], cell[1]) for cell in self.cells]
        for cell in own_cells:
            for a in around(cell[0], cell[1]):
                if a not in own_cells:
                    res.add(a)
        res = list(res)
        res = [a for a in res if 1 <= a[0] <= 10 and 1 <= a[1] <= 10]
        return res

    def __str__(self):
        return str(self.cells)


class SBGameField:
    def __init__(self):
        self.ships = list()
        self.hitten = list()
        self.ship_hitten = list()

        self.xlength = self.ylength = 10

    def clean(self):
        self.ships = list()
        self.hitten = list()
        self.ship_hitten = list()
        self.xlength = self.ylength = 10

    def get_all_ship_cells(self):
        cells = []
        for ship in self.ships:
            for cell in ship.cells:
                cells.append((cell[0], cell[1]))
        return cells

    def __getitem__(self, key):
        key = digit(key)
        rowitems = [None, ] + [EMPTY for cell in range(10)]
        for cell in self.hitten:
            if cell[0] == key:
                rowitems[cell[1]] = cell[2]
        for ship in self.ships:
            for cell in ship.cells:
                if cell[0] == key:
                    rowitems[cell[1]] = cell[2]
        return rowitems

    def field(self):
        return {key: self[key] for key in LETTERKEYS}

    def opfield(self):
        return {key: [(cell if cell != SHIP else EMPTY) for cell in self[key]] for key in LETTERKEYS}

    def __str__(self):
        field = self.field()
        header = '   |  ' + '  |  '.join(key for key in field) + '  |'
        border = '-' * len(header)
        content = '\n'.join(
            ['{:<2}'.format(str(i)) + ' |' + '|'.join(
                '{:^5}'.format(field[key][i]) for key in field.keys()) + '|\n' + border for i in range(1, 11)])
        return header + '\n' + border + '\n' + content

    def oneline(self):
        field = self.field()
        content = ''.join([''.join(field[key][i] for key in field.keys()) for i in range(1, 11)])
        content = ''.join(map(lambda a: a if a is not ' ' else '_', content))
        return content

    def op_oneline(self):
        field = self.opfield()
        content = ''.join([''.join(field[key][i] for key in field.keys()) for i in range(1, 11)])
        content = ''.join(map(lambda a: a if a is not ' ' else '_', content))
        return content

    def as_opposite(self):
        field = self.opfield()
        header = '   |  ' + '  |  '.join(key for key in field) + '  |'
        border = '-' * len(header)
        content = '\n'.join(
            ['{:<2}'.format(str(i)) + ' |' + '|'.join(
                '{:^5}'.format(field[key][i]) for key in field.keys()) + '|\n' + border for i in range(1, 11)])
        return header + '\n' + border + '\n' + content

    def can_ship(self, posx, posy):
        all_cells = self.get_all_ship_cells()
        banned_cells = []
        for cell in all_cells:
            banned_cells += area(cell[0], cell[1])
        return (posx, posy) not in banned_cells

    def put_ship(self, ship):
        crossing = False not in [self.can_ship(cell[0], cell[1]) for cell in ship.cells]
        if not crossing:
            raise ValueError('Здесь нельзя поставить корабль')
        self.ships.append(ship)

    def get_insulted(self):
        return [a for a in self.ship_hitten if a[2] == HIT]

    def hit(self, posx, posy):
        success = False
        res = None
        for ship in self.ships:
            for cell in ship.cells:
                if (cell[0], cell[1]) == (posx, posy):
                    cell[2] = HIT
                    res = HIT
                    self.hitten.append([posx, posy, HIT])
                    success = True
            if ship.isdead():
                for a in ship.around():
                    self.hitten.append([a[0], a[1], MISSED])
                    res = DEAD

                for cell in self.hitten:
                    for c in ship.cells:
                        c[2] = DEAD
                        if (cell[0], cell[1]) == (c[0], c[1]):
                            cell[2] = c[2]

        return res

    def ai_hit(self):
        field = self.opfield()
        hc = map(lambda a: (a[0], a[1]), self.hitten)

        insulted = self.get_insulted()

        if len(insulted) == 0:
            target_chosen = False
            while not target_chosen:
                target = (rnd.randint(1, 10), rnd.randint(1, 10))
                target_chosen = target not in hc
        elif len(insulted) == 1:
            target = rnd.choice(cross_around(insulted[0][0], insulted[0][1]))
        elif len(insulted) > 1:
            xses = [cell[0] for cell in insulted]
            yses = [cell[1] for cell in insulted]
            last = next(a for a in reversed(self.ship_hitten) if a[2] == HIT)
            ca = cross_around(last[0], last[1])
            print(last, ca)
            if len(set(xses)) == 1:
                ca = [a for a in ca if
                      a[0] == xses[0]]
                print(last, ca)
                ca = [a for a in ca if field[letter(a[0])][a[1]] == EMPTY]
                print(last, ca)
                if not ca:
                    last = next(a for a in self.ship_hitten if a[2] == HIT)
                    ca = cross_around(last[0], last[1])
                    print(last, ca)
                    ca = [a for a in ca if
                          a[0] == xses[0]]
                    print(last, ca)
                    ca = [a for a in ca if field[letter(a[0])][a[1]] == EMPTY]
                    print(last, ca)
                try:
                    target = ca[0]
                except IndexError:
                    print(self.as_opposite(), '\n', self)
            elif len(set(yses)) == 1:
                ca = [a for a in ca if
                      a[1] == yses[0]]
                print(last, ca)
                ca = [a for a in ca if field[letter(a[0])][a[1]] == EMPTY]
                print(last, ca)
                if not ca:
                    last = next(a for a in self.ship_hitten if a[2] == HIT)
                    ca = cross_around(last[0], last[1])
                    print(last, ca)
                    ca = [a for a in ca if
                          a[1] == yses[1]]
                    print(last, ca)
                    ca = [a for a in ca if field[letter(a[0])][a[1]] == EMPTY]
                    print(last, ca)
                try:
                    target = ca[0]
                except IndexError:
                    print(self.as_opposite(), '\n', self)

        res = self.hit(*target)

        return res



import random
from pyeasyga import pyeasyga
from ingine import ga

# setup seed data
data = [0, 1, 2, 3, 4, 5, 6, 7]

# initialise the GA
gaa = pyeasyga.GeneticAlgorithm(data,
                                population_size = 200,
                                generations = 100,
                                crossover_probability = 0.8,
                                mutation_probability = 0.2,
                                elitism = True,
                                maximise_fitness = False)


# define and set function to create a candidate solution representation
def create_individual(data):
    individual = data[:]
    random.shuffle(individual)
    return individual


def crossover(parent_1, parent_2):
    crossover_index = random.randrange(1, len(parent_1))
    child_1a = parent_1[:crossover_index]
    child_1b = [i for i in parent_2 if i not in child_1a]
    child_1 = child_1a + child_1b

    child_2a = parent_2[crossover_index:]
    child_2b = [i for i in parent_1 if i not in child_2a]
    child_2 = child_2a + child_2b

    return child_1, child_2


def mutate(individual):
    mutate_index1 = random.randrange(len(individual))
    mutate_index2 = random.randrange(len(individual))
    individual[mutate_index1], individual[mutate_index2] = individual[mutate_index2], individual[mutate_index1]


def selection(population):
    return random.choice(population)


def fitness(individual, data):
    collisions = 0
    for item in individual:
        item_index = individual.index(item)
        for elem in individual:
            elem_index = individual.index(elem)
            if item_index != elem_index:
                if item - (elem_index - item_index) == elem\
                        or (elem_index - item_index) + item == elem:
                    collisions += 1
    return collisions


def print_board(board_representation):
    def print_x_in_row(row_length, x_position):
        print('-' + ''.join(['----' for _ in range(row_length)]))
        print('|' + ''.join([' X |' if i == x_position else '   |' for i in range(row_length)])),

    def print_board_bottom(row_length):
        print('-' + ''.join(['----' for _ in range(row_length)]))

    num_of_rows = len(board_representation)
    row_length = num_of_rows  #rows == columns in a chessboard

    for row in range(num_of_rows):
        print_x_in_row(row_length, board_representation[row])

    print_board_bottom(row_length)
    print('\n')


optimizer = ga.get_optimizer(data,
                             population_size = 200,
                             generations = 100,
                             crossover_probability = 0.8,
                             mutation_probability = 0.2,
                             elitism = True,
                             maximise_fitness = False,
                             create_individual = create_individual,
                             crossover = crossover,
                             mutate = mutate,
                             fitness = fitness,
                             selection = selection)

res = optimizer()

# print the GA's best solution; a solution is valid only if there are no collisions
if res[0] == 0:
    print(res)
    print_board(res[1])
else:
    print(None)



import seabattle_engine as sb
import random as rnd

# инициализация поля
GameField = sb.SBGameField()

for k in range(10000):
    GameField.clean()

    # инициализация кораблей
    for i in range(1, 5):
        for j in reversed(range(5 - i)):
            ship = sb.SBGameShip(length = i, field = GameField)
            ship.randompos()

    for i in range(rnd.randint(1, 130)):
        GameField.ai_hit()

    file = open('data.py', 'a')
    file.write(GameField.op_oneline()+';'+GameField.oneline()+'\n')
    print(k, '%')
    file.close()



from pyeasyga.pyeasyga import GeneticAlgorithm


def get_optimizer(data: list,
                  fitness: type(lambda a: a),
                  population_size: int = 10,
                  generations: int = 100,
                  crossover_probability: float = 0.8,
                  mutation_probability: float = 0.05,
                  elitism: bool = True,
                  maximise_fitness: bool = True,
                  create_individual: type(lambda a: a) = None,
                  crossover: type(lambda a: a) = None,
                  mutate: type(lambda a: a) = None,
                  selection: type(lambda a: a) = None) -> type(lambda a: a):
    """
    :return: ga
        Возвращает объект ga. Методы искать в документации pyeasyga.
    :param data: anything
        Пример существа, генома. Может быть списков.
    :param fitness: function
        Функция приспособленности.
    :param population_size: int, optional
        Размер популяции. По умолчанию равен 10.
    :param generations: int, optional
        Количество поколений. По умолчанию равно 100.
    :param crossover_probability: float, optional
        Вероятность скрещивания. По умолчанию равна 0.8.
    :param mutation_probability: float, optional
        Вероятность мутации. По умолчанию равна 0.05.
    :param elitism: bool, optional
        Применение стратегии элитизма. По умолчанию True.
    :param maximise_fitness: bool, optional
        Ориентирование на увеличение или уменьшение значения фитнесс-функции.
        По умолчанию True.
    :param create_individual: function, optional
        Функция, позволяющая создать случайное существо из data.
        По умолчанию взята из пакета pyeasyga.
    :param crossover: function, optional
        Функция скрещивания.
        По умолчанию взята из пакета pyeasyga.
    :param mutate:
    function, optional
        Функция мутации.
        По умолчанию взята из пакета pyeasyga.
    :param selection: function, optional
        Функция селекции.
        По умолчанию взята из пакета pyeasyga (random selection).
    """

    ga = GeneticAlgorithm(data,
                          population_size = population_size,
                          generations = generations,
                          crossover_probability = crossover_probability,
                          mutation_probability = mutation_probability,
                          elitism = elitism,
                          maximise_fitness = maximise_fitness)
    ga.create_individual = create_individual if create_individual else ga.create_individual
    ga.crossover_function = crossover if crossover else ga.crossover_function
    ga.mutate_function = mutate if mutate else ga.mutate_function
    ga.selection_function = selection if selection else ga.selection_function
    ga.fitness_function = fitness
    ga.run()

    return ga.best_individual



import random as rnd
from ingine import ga


rnd.seed(151685)

points = []

for i in range(22):
    points.append(tuple([rnd.randint(0, 100), rnd.randint(0, 100)]))


def create_individual(points):
    res = points
    rnd.shuffle(res)
    return res


def crossover(creature1, creature2):
    creatures = list(zip(creature1, creature2))
    creature = []
    for _ in creatures:
        choise = rnd.randint(0, 1)
        another = (choise + 1) % 2
        if _[choise] not in creature:
            creature.append(_[choise])
        else:
            creature.append(_[another])
    return creature, list(reversed(creature))


def mutate(creature):
    a = rnd.randint(0, len(creature)-1)
    b = rnd.randint(0, len(creature)-1)
    creature[a], creature[b] = creature[b], creature[a]


def selection(population):
    return rnd.choice(population)


def fitness(creature, data):
    def s(p1, p2):
        return sum([(a - b)**2 for a, b in zip(p1, p2)])

    res = sum([s(prev, current) for prev, current in zip(creature[:-1], creature[1:])])
    res += s(creature[0], creature[-1])

    return res


optimiser = ga.get_optimizer(points,
                             fitness,
                             20,
                             2000,
                             maximise_fitness = False,
                             create_individual = create_individual,
                             mutate = mutate,
                             crossover = crossover)

res = optimiser()[1]

from PIL import Image, ImageDraw

image = Image.new("L", (1000, 1000), (255, ))
draw = ImageDraw.Draw(image)

for x, y in res:
    draw.rectangle((x * 10, y * 10, (x + 1) * 10, (y + 1) * 10), fill = (0,))

coords = []

for prev, cur in zip(res[:-1], res[1:]):
    coords.append(prev[0]*10)
    coords.append(prev[1]*10)
    coords.append(cur[0]*10)
    coords.append(cur[1]*10)

coords.append(res[0][0]*10)
coords.append(res[0][1]*10)
coords.append(res[-1][0]*10)
coords.append(res[-1][1]*10)

draw.line(coords, (0,))

image.show()


import os
import numpy

from keras.datasets import mnist
from keras import utils
from ingine import ann

os.environ['KERAS_BACKEND'] = 'theano'

# Загружаем данные
(X_train, y_train), (X_test, y_test) = mnist.load_data()

# Преобразование размерности изображений
X_train = X_train.reshape(60000, 784)
X_test = X_test.reshape(10000, 784)
# Нормализация данных
X_train = X_train.astype('float32')
X_test = X_test.astype('float32')
X_train /= 255
X_test /= 255

# Преобразуем метки в категории
Y_train = utils.to_categorical(y_train, 10)
Y_test = utils.to_categorical(y_test, 10)

categorizer = ann.get_categorizer(X_train, y_train, 10, epochs = 2)

print(categorizer(X_test[0].reshape(1, 784)))


import numpy

import os

from keras.models import Sequential
from keras.layers import Dense
from keras import utils
from keras import models
import hashlib
import uuid

os.environ['KERAS_BACKEND'] = 'theano'


def hash_data(data):
    """
    Функция хеширования данных. Пригодится для хеширования входных данных для
    последующего сравнения.
    :param data:
    :return: hash
    """
    salt = uuid.uuid4().hex
    return hashlib.sha256(salt.encode() + data.encode()).hexdigest()


def get_categorizer(input_data: numpy.ndarray,
                    result_data: numpy.ndarray,
                    num_cat: int,
                    train_amount: float = 0.8,
                    model: models = Sequential(),
                    layers: list = None,
                    loss: str = "categorical_crossentropy",
                    optimizer: str = "SGD",
                    metrics: str = "accuracy",
                    batch_size: int = None,
                    epochs: int = 1) -> type(lambda a: a):
    """
    Возвращает по полученным параметрам функцию-классификатор
    :param input_data: numpy.ndarray
        Список с входными данными. Каждый элемент списка также может быть списком.
    :param result_data: numpy.ndarray
        Список с ожидаемыми данными. Каждый элемент списка также может быть списком.
    :param num_cat: int
        Количество ожидаемых категорий.
    :param train_amount: float, optional
        Размер обучающей выборки в диапазоне от 0 до 1. По умолчанию 0.8 (80%)
    :param model: keras.models, optional
        Модель из keras.models. По умолчанию Sequential (последовательная)
    :param layers: list, optional
        Список слоев в соответствующем порядке. Слои в списке должны соответствовать
        одному из типов слоёв в keras. По умолчанию используются два полносвязных слоя,
        один из которых по размеру равен количеству свойств объекта, а второй -
        количеству категорий.
    :param loss: string, optional
        Функция ошибки. Должна соответствовать одному из названий функций
        ошибки в keras. По умолчанию = "categorical_crossentrophy"
    :param optimizer: string, optional
        Функция оптимизации. Должна соответствовать одному из названий функций
        оптимизации в keras. По умолчанию = "SGD"
    :param metrics: string, optional
        Функция измерения точности. Должна соответствовать одному из названий метрик
        в keras. По умолчанию = "accuracy"
    :param batch_size: int, optional
        Размер пакета обрабатываемых за один раз обучающих примеров.
        По умолчанию составляет 0.5% от обучающей выборки. Если размер меньше 200,
        он автоматически становится равным 200.
    :param epochs: int, optional
        Количество эпох (циклов обучения на всех элементах выборки). По умолчанию = 1.
    :return: Функция-классификатор. Рекомендуется применять к ней декораторы.
    """
    # Получение отпечатка параметров. Служит для того, чтобы не обучать уже обученную модель.
    params = [a[1] for a in locals().items() if a[0] not in ('fingerprint', 'model')]
    fingerprint = hashlib.md5(str(params).encode()).hexdigest()

    # Попытка загрузить данные о сети, если она уже компилировалась на данном устройстве
    try:
        # Загружаем данные об архитектуре сети
        with open('presaved/%s.json' % fingerprint, "r") as json:
            model = json.read()
        # Создаем модель
        model = models.model_from_json(model)
        # Загружаем сохраненные веса в модель
        model.load_weights('presaved/%s.h5' % fingerprint)
        print("Данные о данной конфигурации сети обнаружены и загружены")
        return lambda a: model.predict(a)
    except FileNotFoundError:
        print("Данные о данной конфигурации сети НЕ обнаружены. Обучение.")

    # Получение размеров пакета и выборок
    idl = len(input_data)
    input_data_test = input_data[int(idl * train_amount):]
    input_data_train = input_data[:int(idl * train_amount)]
    features_num = len(input_data_train[0])
    batch_size = batch_size if batch_size else idl * 0.005
    batch_size = int(batch_size) if batch_size > 200 else 200

    rdl = len(result_data)
    result_data_test = result_data[int(rdl * train_amount):]
    result_data_train = result_data[:int(rdl * train_amount)]
    result_data_test = utils.to_categorical(result_data_test, num_cat)
    result_data_train = utils.to_categorical(result_data_train, num_cat)

    # Инициализация слоёв
    if layers:
        for layer in layers:
            model.add(layer)
    else:
        model.add(Dense(features_num, input_dim = features_num, activation = 'relu', kernel_initializer = "normal"))
        model.add(Dense(num_cat, input_dim = num_cat, activation = "softmax", kernel_initializer = "normal"))

    # Компиляция модели
    model.compile(loss = loss, optimizer = optimizer, metrics = [metrics])
    model.fit(input_data_train, result_data_train, batch_size = batch_size, epochs = epochs)

    # Вывод метрик
    scores = model.evaluate(input_data_test, result_data_test, verbose = 1)
    print(scores)

    # сохранение модели
    try:
        os.mkdir('presaved')
    except FileExistsError:
        pass
    with open('presaved/%s.json' % fingerprint, 'w') as f:
        f.write(model.to_json())
    model.save_weights('presaved/%s.h5' % fingerprint)

    return lambda a: model.predict(a)


def get_regression(input_data: numpy.ndarray,
                   result_data: numpy.ndarray,
                   train_amount: float = 0.8,
                   model: models = Sequential(),
                   layers: list = None,
                   loss: str = "mse",
                   optimizer: str = "adam",
                   metrics: str = "mse",
                   batch_size: int = None,
                   epochs: int = 1) -> type(lambda a: a):
    """
    Возвращает по полученным параметрам функцию-классификатор
    :param input_data: numpy.ndarray
        Список с входными данными. Каждый элемент списка также может быть списком.
    :param result_data: numpy.ndarray
        Список с ожидаемыми данными. Каждый элемент списка также может быть списком.
    :param train_amount: float, optional
        Размер обучающей выборки в диапазоне от 0 до 1. По умолчанию 0.8 (80%)
    :param model: keras.models, optional
        Модель из keras.models. По умолчанию Sequential (последовательная)
    :param layers: list, optional
        Список слоев в соответствующем порядке. Слои в списке должны соответствовать
        одному из типов слоёв в keras. По умолчанию используются два полносвязных слоя,
        один из которых по размеру равен количеству свойств объекта, а второй -
        выходному значению.
    :param loss: string, optional
        Функция ошибки. Должна соответствовать одному из названий функций
        ошибки в keras. По умолчанию = "mse" - среднеквадратичная
    :param optimizer: string, optional
        Функция оптимизации. Должна соответствовать одному из названий функций
        оптимизации в keras. По умолчанию = "adam"
    :param metrics: string, optional
        Функция измерения точности. Должна соответствовать одному из названий метрик
        в keras. По умолчанию = "mse" - среднеквадратичная
    :param batch_size: int, optional
        Размер пакета обрабатываемых за один раз обучающих примеров.
        По умолчанию составляет 0.5% от обучающей выборки. Если размер меньше 200,
        он автоматически становится равным 200.
    :param epochs: int, optional
        Количество эпох (циклов обучения на всех элементах выборки). По умолчанию = 1.
    :return: Функция-классификатор. Рекомендуется применять к ней декораторы.
    """
    # Получение отпечатка параметров. Служит для того, чтобы не обучать уже обученную модель.
    params = [a[1] for a in locals().items() if a[0] not in ('fingerprint', 'model')]
    fingerprint = hashlib.md5(str(params).encode()).hexdigest()

    # Попытка загрузить данные о сети, если она уже компилировалась на данном устройстве
    try:
        # Загружаем данные об архитектуре сети
        with open('presaved/%s.json' % fingerprint, "r") as json:
            model = json.read()
        # Создаем модель
        model = models.model_from_json(model)
        # Загружаем сохраненные веса в модель
        model.load_weights('presaved/%s.h5' % fingerprint)
        print("Данные о данной конфигурации сети обнаружены и загружены")
        return lambda a: model.predict(a)
    except FileNotFoundError:
        print("Данные о данной конфигурации сети НЕ обнаружены. Обучение.")

    # Получение размеров пакета и выборок
    idl = len(input_data)
    input_data_test = input_data[int(idl * train_amount):]
    input_data_train = input_data[:int(idl * train_amount)]
    features_num = len(input_data_train[0])
    batch_size = batch_size if batch_size else idl * 0.005
    batch_size = int(batch_size) if batch_size > 200 else 200

    rdl = len(result_data)
    result_data_test = result_data[int(rdl * train_amount):]
    result_data_train = result_data[:int(rdl * train_amount)]

    # Инициализация слоёв
    if layers:
        for layer in layers:
            model.add(layer)
    else:
        model.add(Dense((features_num - 1) ** 2, input_dim = features_num, activation = 'relu',
                        kernel_initializer = "normal"))
        model.add(Dense(1))

    # Компиляция модели
    model.compile(loss = loss, optimizer = optimizer, metrics = [metrics])
    model.fit(input_data_train, result_data_train, batch_size = batch_size, epochs = epochs)

    # Вывод метрик
    scores = model.evaluate(input_data_test, result_data_test, verbose = 1)
    print(scores)

    # сохранение модели
    try:
        os.mkdir('presaved')
    except FileExistsError:
        pass
    with open('presaved/%s.json' % fingerprint, 'w') as f:
        f.write(model.to_json())
    model.save_weights('presaved/%s.h5' % fingerprint)

    return lambda a: model.predict(a)


def get_customnn(input_data: numpy.ndarray,
                 result_data: numpy.ndarray,
                 layers: list,
                 train_amount: float = 0.8,
                 model: models = Sequential(),
                 loss: str = "mae",
                 optimizer: str = "SGD",
                 metrics: str = "mse",
                 batch_size: int = None,
                 epochs: int = 1) -> type(lambda a: a):
    """
    Возвращает по полученным параметрам функцию нейронной сети с пользовательскими слоями.
    :param input_data: numpy.ndarray
        Список с входными данными. Каждый элемент списка также может быть списком.
    :param result_data: numpy.ndarray
        Список с ожидаемыми данными. Каждый элемент списка также может быть списком.
    :param layers: list
        Список слоев в соответствующем порядке. Слои в списке должны соответствовать
        одному из типов слоёв в keras.
    :param train_amount: float, optional
        Размер обучающей выборки в диапазоне от 0 до 1. По умолчанию 0.8 (80%)
    :param model: keras.models, optional
        Модель из keras.models. По умолчанию Sequential (последовательная)
    :param loss: string, optional
        Функция ошибки. Должна соответствовать одному из названий функций
        ошибки в keras. По умолчанию = "mae"
    :param optimizer: string, optional
        Функция оптимизации. Должна соответствовать одному из названий функций
        оптимизации в keras. По умолчанию = "SGD"
    :param metrics: string, optional
        Функция измерения точности. Должна соответствовать одному из названий метрик
        в keras. По умолчанию = "mse"
    :param batch_size: int, optional
        Размер пакета обрабатываемых за один раз обучающих примеров.
        По умолчанию составляет 0.5% от обучающей выборки. Если размер меньше 200,
        он автоматически становится равным 200.
    :param epochs: int, optional
        Количество эпох (циклов обучения на всех элементах выборки). По умолчанию = 1.
    :return: Функция-классификатор. Рекомендуется применять к ней декораторы.
    """
    # Получение отпечатка параметров. Служит для того, чтобы не обучать уже обученную модель.
    params = [a[1] for a in locals().items() if a[0] not in ('fingerprint', 'model')]
    print(params)
    fingerprint = hashlib.md5(str(params).encode()).hexdigest()

    # Попытка загрузить данные о сети, если она уже компилировалась на данном устройстве
    try:
        # Загружаем данные об архитектуре сети
        with open('presaved/%s.json' % fingerprint, "r") as json:
            model = json.read()
        # Создаем модель
        model = models.model_from_json(model)
        # Загружаем сохраненные веса в модель
        model.load_weights('presaved/%s.h5' % fingerprint)
        print("Данные о данной конфигурации сети обнаружены и загружены")
        return lambda a: model.predict(a)
    except FileNotFoundError:
        print("Данные о данной конфигурации сети НЕ обнаружены. Обучение.")

    # Получение размеров пакета и выборок
    idl = len(input_data)
    input_data_test = input_data[int(idl * train_amount):]
    input_data_train = input_data[:int(idl * train_amount)]
    features_num = len(input_data_train[0])
    batch_size = batch_size if batch_size else idl * 0.005
    batch_size = int(batch_size) if batch_size > 200 else 200

    rdl = len(result_data)
    result_data_test = result_data[int(rdl * train_amount):]
    result_data_train = result_data[:int(rdl * train_amount)]

    for layer in layers:
        model.add(layer)

    # Компиляция модели
    model.compile(loss = loss, optimizer = optimizer, metrics = [metrics])
    model.fit(input_data_train, result_data_train, batch_size = batch_size, epochs = epochs)

    # Вывод метрик
    scores = model.evaluate(input_data_test, result_data_test, verbose = 1)
    print(scores)

    # сохранение модели
    try:
        os.mkdir('presaved')
    except FileExistsError:
        pass
    with open('presaved/%s.json' % fingerprint, 'w') as f:
        f.write(model.to_json())
    model.save_weights('presaved/%s.h5' % fingerprint)

    return lambda a: model.predict(a)



from setuptools import setup, find_packages

import ingine

with open("README.md", "r") as fh:
    long_description = fh.read()

print(find_packages())

setup(
    name = "ingine",
    version = ingine.__version__,
    author = "sqarrt",
    author_email = "sqarrt1337@gmail.com",
    description = "A cross-platform library for easy-access AI tools",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/sqarrt/ingine",
    packages = find_packages(),
    install_requires = ['keras',
                        'numpy',
                        'tensorflow',
                        'theano',
                        'pyeasyga'],
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: Russian",
    ],
)



import numpy

import os

os.environ['KERAS_BACKEND'] = 'theano'

from keras.layers import Dense
import csv

from ingine import ann

# Устанавливаем seed для повторяемости результатов
numpy.random.seed(151681)


def load_data():
    """
    Функция получения данных из data.py
    :return: list(2) of lists(2)
    """
    input_data = []
    output_data = []

    from data import data

    data = [_.split(';') for _ in data.split('\n')]

    for row in data:
        in_data = row[0]
        out_data = row[1]
        in_data = in_data.replace('X', '1')
        in_data = in_data.replace('+', '1')
        in_data = in_data.replace('_', '0')
        in_data = in_data.replace('-', '0')

        out_data = out_data.replace('X', '0')
        out_data = out_data.replace('+', '0')
        out_data = out_data.replace('O', '1')
        out_data = out_data.replace('_', '0')
        out_data = out_data.replace('-', '0')

        in_data = list(map(lambda a: int(a), in_data))
        out_data = list(map(lambda a: int(a), out_data))

        input_data.append(in_data)
        output_data.append(out_data)

    input_data = numpy.asarray(input_data)
    output_data = numpy.asarray(output_data)

    input_data_test = input_data[8000:]
    input_data = input_data[:8000]

    output_data_test = output_data[8000:]
    output_data = output_data[:8000]

    return (input_data, output_data), (input_data_test, output_data_test)


# Загружаем данные
(X_train, Y_train), (X_test, Y_test) = load_data()

# Определеяем слои
layers = [Dense(100, input_dim = 100, activation = "softsign", kernel_initializer = "normal"),
          Dense(20, activation = "softsign", kernel_initializer = "normal"),
          Dense(10, activation = "softsign", kernel_initializer = "normal"),
          Dense(100, activation = "softsign", kernel_initializer = "normal")]

# Получение модели с
customnn = ann.get_customnn(X_train, Y_train, layers = layers, epochs = 5)

res = customnn(X_test)

xt = numpy.asarray([255 if a == 1 else a for a in X_test[15]]).reshape(10, 10)
yt = numpy.asarray([255 if a == 1 else a for a in Y_test[15]]).reshape(10, 10)
rt = numpy.asarray([int(-numpy.sign(a)) * 255 for a in res[15]]).reshape(10, 10)

from PIL import Image, ImageDraw

image = Image.new("L", (500, 500), (255,))
draw = ImageDraw.Draw(image)
for i in range(10):  # For every pixel:
    for j in range(10):
        draw.rectangle((i * 50, j * 50, (i + 1) * 50, (j + 1) * 50), fill = (xt[i][j],))

image.save("images/xt.png", "PNG")

image = Image.new("L", (500, 500), (255,))
draw = ImageDraw.Draw(image)
for i in range(10):  # For every pixel:
    for j in range(10):
        draw.rectangle((i * 50, j * 50, (i + 1) * 50, (j + 1) * 50), fill = (yt[i][j],))

image.save("images/yt.png", "PNG")

image = Image.new("L", (500, 500), (255,))
draw = ImageDraw.Draw(image)
for i in range(10):  # For every pixel:
    for j in range(10):
        draw.rectangle((i * 50, j * 50, (i + 1) * 50, (j + 1) * 50), fill = (rt[i][j],))

image.save("images/rt.png", "PNG")



import os
import numpy

from keras.datasets import boston_housing
from keras import utils
from ingine import ann

os.environ['KERAS_BACKEND'] = 'theano'

# Загружаем данные
(x_train, y_train), (x_test, y_test) = boston_housing.load_data()

# Среднее значение
mean = x_train.mean(axis=0)
# Стандартное отклонение
std = x_train.std(axis=0)
x_train -= mean
x_train /= std
x_test -= mean
x_test /= std

regression = ann.get_regression(x_train, y_train, epochs = 101, loss = 'mse')

print(regression(x_test[50].reshape(1, 13))[0], y_test[50])